#ifndef __CHANGE_SET_HH__
#define __CHANGE_SET_HH__

// copyright (C) 2004 graydon hoare <graydon@pobox.com>
// all rights reserved.
// licensed to the public under the terms of the GNU GPL (>= 2)
// see the file COPYING for details

#include <map>
#include <string>

#include <boost/shared_ptr.hpp>

#include "interner.hh"
#include "manifest.hh"
#include "vocab.hh"

// a change_set is a text object. It has a precise, normalizable serial form
// as UTF-8 text. it also has some sub-components. not all of these
// sub-components are separately serialized (they could be but there is no
// call for it). a grammar (aside from the parsing code) for the serialized
// form will show up here eventually. until then, here is an example.
//
// change_set:
// {
//   paths:
//   {
//      rename_file:
//      {
//        src: "usr/bin/sh"
//        dst: "usr/bin/foo"
//      }
//      delete_dir: "usr/bin"
//      add_file: "tmp/foo/bar.txt"
//   }
//   deltas:
//   {
//     delta:
//     {
//       path: "tmp/foo/bar.txt"
//       src: [71e0274f16cd68bdf9a2bf5743b86fcc1e597cdc]     
//       dst: [71e0274f16cd68bdf9a2bf5743b86fcc1e597cdc]
//     }
//   }
// } 
//
// note that this object is made up of two important sub-components: 
// path_edits and deltas. "path_edits" is exactly the same object stored
// in MT/path_edits (formerly MT/work). 

struct
change_set
{  
  struct
  path_rearrangement
  {
    std::set<file_path> deleted_files;
    std::set<file_path> deleted_dirs;
    std::map<file_path, file_path> renamed_files;
    std::map<file_path, file_path> renamed_dirs;
    std::set<file_path> added_files;
    bool operator==(path_rearrangement const & other) const;
    bool empty() const;
  };

  typedef std::map<file_path, std::pair<file_id, file_id> > delta_map;
  
  path_rearrangement rearrangement;
  delta_map deltas;

  bool operator==(change_set const & other) const;
  bool empty() const;
  void add_file(file_path const & a);
  void add_file(file_path const & a, file_id const & ident);
  void apply_delta(file_path const & path, 
		   file_id const & src, 
		   file_id const & dst);
  void delete_file(file_path const & d);
  void delete_dir(file_path const & d);
  void rename_file(file_path const & a, file_path const & b);
  void rename_dir(file_path const & a, file_path const & b);
};

inline bool 
null_name(file_path const & p)
{
  return p().empty();
}

inline bool 
null_id(file_id const & i)
{
  return i.inner()().empty();
}

inline file_path const & 
delta_entry_path(change_set::delta_map::const_iterator i)
{
  return i->first;
}

inline file_id const & 
delta_entry_src(change_set::delta_map::const_iterator i)
{
  return i->second.first;
}

inline file_id const & 
delta_entry_dst(change_set::delta_map::const_iterator i)
{
  return i->second.second;
}

void
apply_rearrangement_to_filesystem(change_set::path_rearrangement const & re,
				  local_path const & temporary_root);


// merging and concatenating 

void
normalize_change_set(change_set & n);

void
concatenate_change_sets(change_set const & a,
			change_set const & b,
			change_set & concatenated);

struct merge_provider;

void
merge_change_sets(change_set const & a,
		  change_set const & b,
		  change_set & a_merged,
		  change_set & b_merged,
		  merge_provider & merger,
		  app_state & app);

// value-oriented access to printers and parsers

void
read_path_rearrangement(data const & dat,
			change_set::path_rearrangement & re);

void
write_path_rearrangement(change_set::path_rearrangement const & re,
			 data & dat);

void
read_change_set(data const & dat,
		change_set & cs);

void
write_change_set(change_set const & cs,
		 data & dat);

void
apply_path_rearrangement(path_set const & old_ps,
			 change_set::path_rearrangement const & pr,
			 path_set & new_ps);

void
apply_path_rearrangement(manifest_map const & old_m,
			 change_set::path_rearrangement const & pr,
			 manifest_map & old_m_rearranged);

void
build_pure_addition_change_set(manifest_map const & man,
			       change_set & cs);

void
apply_change_set(manifest_map const & old_man,
		 change_set const & cs,
		 manifest_map & new_man);

// utility for log walker
file_path
apply_change_set_inverse(change_set const & cs,
			 file_path const & file_in_second);

// quick, optimistic and destructive version for rcs importer
void
apply_change_set(change_set const & cs,
		 manifest_map & man);

void 
invert_change_set(change_set const & a2b,
		  manifest_map const & a_map,
		  change_set & b2a);


// basic_io access to printers and parsers

namespace basic_io { struct printer; struct parser; }

void 
print_change_set(basic_io::printer & printer,
		 change_set const & cs);

void 
parse_change_set(basic_io::parser & parser,
		 change_set & cs);

void 
print_path_rearrangement(basic_io::printer & basic_printer,
			 change_set::path_rearrangement const & pr);

void 
parse_path_rearrangement(basic_io::parser & pa,
			 change_set::path_rearrangement & pr);

#endif // __CHANGE_SET_HH__
